;;; -*- auto-recompile: t -*-

;;; csound-sco.el --- major mode for editing Csound scores

;;; This file is not part of GNU Emacs.

;;; This program is free software; you can redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation; either version 2, or (at your option)
;;; any later version.

;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.

;;; You should have received a copy of the GNU General Public License
;;; along with GNU Emacs; see the file COPYING.  If not, write to
;;; the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

; ---------------
; initially based on the code by John Fitch (26 Oct 1996), version 1.05
; heavily hacked and adapted for Csound-X
; by Stef <hepta@zogotounga.net>

; last modification on April 4, 2006

(require 'csound-orc)
(require 'csound-mx)

(defgroup csound-sco nil
  "csound-sco major mode settings"
  :group 'csound-x
  :prefix "csound-sco-")

(defcustom csound-sco-mode-never-indent nil
  "If t, the indentation code for csound-sco-mode is turned off."
  :type 'boolean
  :group 'csound-sco)

(defcustom csound-sco-mode-hook nil
  "Hook run when csound-sco-mode is started."
  :type 'hook
  :group 'csound-sco)

(defcustom csound-sco-mode-load-hook nil
  "Hook run when csound-sco-mode is loaded."
  :type 'hook
  :group 'csound-sco)


;;======================================================================
;                       csound-sco-mode
;;======================================================================

(defun csound-sco-mode ()
  "Mode for editing Csound Scores

\\{csound-sco-mode-map}"
  (interactive)
  (kill-all-local-variables)
  (use-local-map csound-sco-mode-map)
  (set-syntax-table csound-sco-mode-syntax-table)
  (setq mode-name "Csound Score")
  (setq major-mode 'csound-sco-mode)
  (make-local-variable 'comment-start)
  (make-local-variable 'comment-end)
  (make-local-variable 'comment-column)
  (make-local-variable 'indent-line-function)
  (setq comment-start "; "
	comment-end ""
	comment-column 40
	indent-line-function 'csound-sco-indent-line)
  (make-local-variable 'comment-indent-function)
  (setq comment-indent-function 'csound-sco-mode-comment)
  (make-local-variable 'comment-start-skip)
  (setq comment-start-skip ";+[ \t]*")
  (setq csound-sco-inline-empty-comment-pattern "^.+;+ *$")
  (setq csound-sco-code-level-empty-comment-pattern "^[\t ]+;; *$")
  (setq csound-sco-flush-left-empty-comment-pattern "^;;; *$")
  (make-local-variable 'indent-tabs-mode)
  (setq indent-tabs-mode nil)
  ;; multiline comments:
  (make-variable-buffer-local font-lock-fontify-region-function)
  (setq font-lock-fontify-region-function 'csound-scorc-fontify-region)
  ;; font lock
  (make-local-variable 'font-lock-defaults)
  (setq font-lock-defaults '(csound-sco-font-lock-keywords
			     nil nil nil beginning-of-defun
			     (font-lock-mark-block-function mark-defun)))
  ;; menu
  (csound-install-sco-menu)
  (run-hooks 'csound-sco-mode-hook)
;  (font-lock-mode 1)
  ;; matrices:
  (if (featurep 'csound-mx)
      (csound-mx-install)
      (scomx-highlight-all)))

(defvar csound-sco-mode-map (make-sparse-keymap) "Keymap for csound-sco-mode")

(if csound-x-disable-smart-semicolon
    (define-key csound-sco-mode-map ";" nil)
  (define-key csound-sco-mode-map ";" 'csound-sco-comment))
(define-key csound-sco-mode-map "\C-c;"    'comment-region)
(define-key csound-sco-mode-map "\C-c:"    'uncomment-region)
(define-key csound-sco-mode-map [(control c) (control ?2)] 'csound-e-power) 

;; indentation keys 
(define-key csound-sco-mode-map "\t" 'csound-sco-mode-indent-command)
(define-key csound-sco-mode-map "\C-j" 'newline-and-indent)

(defvar csound-sco-mode-syntax-table nil "Syntax table")

(unless csound-sco-mode-syntax-table
  (setq csound-sco-mode-syntax-table
	(make-syntax-table text-mode-syntax-table)))

;;======================================================================
;                       indentation
;;======================================================================

(defun csound-sco-mode-indent-command ()
  "Command for indenting text. 
Just calls csound-sco-mode-indent."
  (interactive)
  (csound-sco-mode-indent))

(defun csound-sco-mode-indent ()
  "indentation workhorse function."
  (unless csound-sco-mode-never-indent

    (save-excursion
      (goto-char (point-at-bol))
      (save-restriction
        (narrow-to-region (point-at-bol) (point-at-eol))
        (delete-horizontal-space)
        (cond ((looking-at "[iftsarb0-9.-]")  
               (let* ((look-up (csound-sco-look-up))
                      (tab-stop-list
                       (cond
                        ((string= look-up "i")
                         '(2 4 7 14 21 28 35 42 49 56 63 70 77 84 91 98 105 112))
                        ((string= look-up "f")
                         '(2 4 6 12 18 24 30 36 42 48 54 60 66 72 78 84 90 96 102 108 114))
                        (t tab-stop-list) ;;  continuer...
                        )))
                 (when (looking-at "[0-9.-]")
                   (insert (make-string (third tab-stop-list) ? )))
                 (while (and (re-search-forward "[ \t]+" (point-max) t)
                             (not (looking-at "$"))
                             (not (looking-at "[;\]")))
                   (forward-char -1)
                   (csound-shrink-horizontal-space)
                   (csound-ensure-tab-stop))
                 (csound-tabulate-left-comment)))
              ; comment line
              ((looking-at "[ \t]*;")
               (delete-horizontal-space))
              ; end
              ((looking-at "[ \t]e")
               (delete-horizontal-space)
               (csound-tabulate-left-comment)))))))

(defun csound-sco-look-up ()
  (save-restriction
    (widen)
    (save-excursion
      (goto-char (point-at-bol))
      (skip-chars-forward " \t")
      (if (not (looking-at "[0-9-]"))
          (buffer-substring-no-properties (point) (1+ (point)))
        (forward-line -1)
        (csound-sco-look-up)))))

(defun csound-ensure-tab-stop ()
  (when (> (save-excursion (backward-char 1) (move-to-tab-stop) (point))
           (point))
      (tab-to-tab-stop)))

(defun csound-sco-mode-comment ()
  (save-excursion
    (beginning-of-line)
    (if (looking-at ";")
	0
      (skip-chars-backward " \t")
      (max (if (bolp) 0 (1+ (current-column)))
	   comment-column))))

(defun csound-sco-indent-line ()
  "Indents current Csound line based on its contents."
  (interactive)
  (csound-sco-mode-indent))

(defun csound-sco-indent-comment ()
  "Align or create comment on current line.
Existing comments of all types are recognized and aligned.
If the line has no comment, a side-by-side comment is inserted and aligned
if the value of  comment-start  is not nil.
Otherwise, a separate-line comment is inserted, on this line
or on a new line inserted before this line if this line is not blank."
  (interactive)
  (beginning-of-line)
  ;; Recognize existing comments
  (cond ((looking-at ";")
	 (csound-sco-indent-line))
        ; otherwise goto end of line or sth else?
	;; No existing comment.
	;; If side-by-side comments are defined, insert one,
	;; unless line is now blank.
	((and comment-start (not (looking-at "^[ \t]*$")))
	 (end-of-line)
	 (delete-horizontal-space)
	 (indent-to (csound-sco-mode-comment))
	 (insert comment-start))
	;; Else insert separate-line comment, making a new line if nec.
	(t
	 (if (looking-at "^[ \t]*$")
	     (delete-horizontal-space)
;;	   (beginning-of-line)
;;	   (insert "\n")
;;	   (forward-char -1)
	   )
	 (insert ";")
	 (insert-char '\; (- (calculate-csound-sco-indent) (current-column))))))

(defun csound-sco-find-comment-start-skip ()
  "Move to past `comment-start-skip' found on current line.
Return t if `comment-start-skip' found, nil if not."
;;; In order to move point only if comment-start-skip is found,
;;; this one uses a lot of save-excursions.  Note that re-search-forward
;;; moves point even if comment-start-skip is inside a string-constant.
;;; Some code expects certain values for match-beginning and end
  (interactive)
  (if (save-excursion
	(re-search-forward comment-start-skip
			   (save-excursion (end-of-line) (point)) t))
      (let ((save-match-beginning (match-beginning 0))
	    (save-match-end (match-end 0)))
	(goto-char save-match-beginning)
	(re-search-forward comment-start-skip
			   (save-excursion (end-of-line) (point)) t)
	(goto-char (match-end 0))
	t)
    nil))

(defun csound-sco-current-line-indentation ()
  "Indentation of current line
This is the column position of the first non-whitespace character"
  (save-excursion
    (beginning-of-line)
    (cond ((looking-at "^;")
	   (goto-char (match-end 0))
	   (skip-chars-forward ";")))
    ;; Move past whitespace.
    (skip-chars-forward " \t")
    (current-column)))

(defun calculate-csound-sco-indent ()
  "Calculates the Csound indent column based on previous lines."
  (let (icol)
    (save-excursion
      (if (= (point) (point-min))
	  (setq icol 8)
	(save-excursion
	  (beginning-of-line)
	  (cond ((looking-at "^;")
		 (setq icol 0))
	  ;; Move past whitespace.
		(t (skip-chars-forward " \t")
		   (setq icol (current-column))))))
      (save-excursion
	(beginning-of-line)
	icol))))

;;======================================================================
;                       fontlock support
;;======================================================================

(dolist 
    (fspec
     '( ;; name  foreground  background  bold doc
       (csound-table-face "RoyalBlue" nil nil "f-tables")
       (csound-tableno-face "red" nil t "f-tables GEN type")
       (csound-i1-face "Sienna"	nil nil "i..1 statements")
       (csound-i2-face "DarkSlateBlue" "gray91" nil "i..1 statements")
       (csound-i3-face "SteelBlue" "gray92" nil "i..2 statements")
       (csound-i4-face "Brown" "gray93" nil "i..3 statements")
       (csound-i5-face "DarkOliveGreen" "gray94" nil "i..4 statements")
       (csound-i6-face "MediumSeaGreen" "gray95" nil "i..5 statements")
       (csound-i7-face "Maroon"	"gray96" nil "i..6 statements")
       (csound-i8-face "Plum" "gray97" nil "i..7 statements")
       (csound-i9-face "OrangeRed" "gray98" nil "i..8 statements")
       (csound-i0-face "Violet"	"gray99" nil "i..0 statements")
       (csound-section-face "Sienna" nil t "section statements")
       (csound-tempo-face "ForestGreen"	nil nil "tempo statements")))
(eval
 `(defface ,(nth 0 fspec) 
    '((t (:foreground ,(nth 1 fspec) 
          :background ,(nth 2 fspec)
          :bold ,(nth 3 fspec))))
    ,(format "Face for %s in csound scores" (nth 4 fspec))
    :group 'csound-faces)))

(defvar csound-sco-font-lock-keywords 
  (list
   ;; Comments
   '(";.*$" . 'csound-comment-face)
   ;;
   ;; Fontify syntactically (assuming strings cannot be quoted or span lines).
   '("\"[^\"\n]*\"" . 'csound-string-face)
   ;;
   '("^[ \t]*[es]" . 'csound-section-face)
   ;; f-tables
   '("^[ \t]*\\(f[ \t]*[0-9]+[ \t]+[0-9.]+[ \t]+[0-9]+\\)[ \t]+\\(-?[0-9]+\\)[ \t]+\\(.*\\)"
     (1 'csound-table-face)
     (2 'csound-tableno-face)
     (3 'csound-table-face keep t))
   ;;    
   '("^[ \t]*i[ \t]*[0-9]*1\\b.*" . 'csound-i1-face)
   '("^[ \t]*i[ \t]*[0-9]*2\\b.*" . 'csound-i2-face)
   '("^[ \t]*i[ \t]*[0-9]*3\\b.*" . 'csound-i3-face)
   '("^[ \t]*i[ \t]*[0-9]*4\\b.*" . 'csound-i4-face)
   '("^[ \t]*i[ \t]*[0-9]*5\\b.*" . 'csound-i5-face)
   '("^[ \t]*i[ \t]*[0-9]*6\\b.*" . 'csound-i6-face)
   '("^[ \t]*i[ \t]*[0-9]*7\\b.*" . 'csound-i7-face)
   '("^[ \t]*i[ \t]*[0-9]*8\\b.*" . 'csound-i8-face)
   '("^[ \t]*i[ \t]*[0-9]*9\\b.*" . 'csound-i9-face)
   '("^[ \t]*i[ \t]*[0-9]*0\\b.*" . 'csound-i0-face)
   '("^[ \t]*\\bs[0-9]*\\b.*" . 'csound-krate-face)
   '("^[ \t]*\\b[btv][0-9]*\\b.*" . 'csound-tempo-face)
   ;;
;   '("\\b<\\|\\b>\\b\\|\\b{\\|\\b}\\b" . 'font-lock-string)
;   '("\\b+\\b" . 'font-lock-global)
;   '("\\b\\.\\b" . 'csound-tempo-face)
   ;;
   ;; #include & #define
   '("#" 0 'csound-define-face t)
   '("^[ \t]*#define\\([^(#]*\\)" 1 'csound-flow-face t)
   '("^[ \t]*#include" 0 'csound-define-face t)
   '("^[ \t]*#define" 0 'csound-define-face t)
   ;;
   ;; Macros
   '("\\$[a-zA-Z0-9_]+" 0 'csound-flow-face t)  
   ;;
   ;; Comments ; & /* */
   '("/\\*\\(.*\\)\\*/" 1 'csound-comment-face t)
   '("\\(;[^|\n\r].*$\\)" 1 'csound-comment-face t)
   ;;
   ;; Meta-labels
   '(";|.*$" 0 'csound-label-face t)
   '(";_.*_[be]\\.[0-9]+" 0 'csound-label-face t)
   )
  "Default expressions to highlight in Csound mode.")

;;======================================================================
;                       commenting/uncommenting
;;======================================================================

(defvar csound-sco-code-level-empty-comment-pattern nil
  "Pattern for entry line with comment")
(defvar csound-sco-flush-left-empty-comment-pattern nil
  "Pattern for left most comment")
(defvar csound-sco-inline-empty-comment-pattern nil
  "Patern for comment line inline")

(defun csound-sco-line-matches (pattern &optional withcomment)
  (save-excursion
    (beginning-of-line)
    (looking-at pattern)))

(defun csound-sco-pop-comment-level ()
  ;; Delete an empty comment ending current line.  Then set up for a new one,
  ;; on the current line if it was all comment, otherwise above it
  (end-of-line)
  (delete-horizontal-space)
  (while (= (preceding-char) ?; )
    (delete-backward-char 1))
  (delete-horizontal-space)
  (if (bolp)
      nil
    (beginning-of-line)
    (open-line 1)))

(defun csound-sco-comment ()
  "Convert an empty comment to a `larger' kind, or start a new one.
These are the known comment classes:

   1 -- comment to the right of the code (at the comment-column)
   2 -- comment on its own line, indented like code
   3 -- comment on its own line, beginning at the left-most column.

Suggested usage:  while writing your code, trigger csound-sco-comment
repeatedly until you are satisfied with the kind of comment."
  (interactive)
  (cond

   ;; Blank line?  Then start comment at code indent level.
   ((csound-sco-line-matches "^[ \t]*$")
    (delete-horizontal-space)
    (tab-to-tab-stop)
    (insert comment-start))

   ;; Nonblank line with no comment chars in it?
   ;; Then start a comment at the current comment column
   ((csound-sco-line-matches "^[^;\n]+$")
    (indent-for-comment))

   ;; Flush-left comment present?  Just insert character.
   ((csound-sco-line-matches csound-sco-flush-left-empty-comment-pattern)
    (insert ";"))

   ;; Empty code-level comment already present?
   ;; Then start flush-left comment, on line above if this one is nonempty. 
   ((csound-sco-line-matches csound-sco-code-level-empty-comment-pattern)
    (csound-sco-pop-comment-level)
    (insert ";;;"))

   ;; Empty comment ends line?
   ;; Then make code-level comment, on line above if this one is nonempty. 
   ((csound-sco-line-matches csound-sco-inline-empty-comment-pattern)
    (csound-sco-pop-comment-level)
    (tab-to-tab-stop)
    (insert ";; "))

   ;; If all else fails, insert character
   (t
    (insert ";")))
  (end-of-line))

(defun uncomment-region ()
  "Call `\\[comment-region]' with an argument"
  (interactive)
  (comment-region (point) (mark) '(t)))

;;======================================================================
;                       menu
;;======================================================================

(defun csound-install-sco-menu ()
  (easy-menu-define csound-sco-menu csound-sco-mode-map
    "Menu provided by `csound-sco-mode'"
    `("Csound/Sco"
      ["Comment Region" comment-region t]
      ["Uncomment Region" uncomment-region t]
      ["Power-of-2 Region" csound-e-power t]
      "--"
      ["Wrap in .csd"           cscsd-wrap-buffer (and (cscsd-non-indirect-buffer-p)
						       (not (cscsd-buffer-is-a-csd-p)))]
      ["Send to muO"  muo-get-sco (featurep 'surmulot) ]
      ["Visit associated orchestra" (find-file (funcall cscsd-associated-orc)) (funcall cscsd-associated-orc)]
      "--"
      ;; the repository submenu is only computed upon request:
      ,(or cscsd-repository-menu-cache (cscsd-make-repository-menu))
      "--"
      ["Display f-table at point"   cscsd-display-table (thing-at-point-looking-at "^[ \t]*f.*")]
      ["Source of macro at point"        cscsd-source-macro 
       (thing-at-point-looking-at "\\$[a-zA-Z0-9_]+")]
      ["Visit included file at point"    cscsd-find-include  
       (thing-at-point-looking-at "^[ \t]*#include.*")]
      "--"
      ["Electric <return>" (setq cscsd-sco-has-electric-return 
                            (not cscsd-sco-has-electric-return))
       :style toggle
       :selected cscsd-sco-has-electric-return]
      "--"
      ,(cscsd-processing-menu)
      ["Process source CSD"               (cscsd-process (buffer-base-buffer (current-buffer))) (cscsd-indirect-buffer-p)]
      ["Process via temp CSD" cscsd-process-scorc-via-temp-csd
       (and (cscsd-non-indirect-buffer-p) (not (cscsd-buffer-is-a-csd-p)))]
      "--"
      ["Compile score (to DAC)"          cscsd-play-score (null (cscsd-buffer-is-a-csd-p))]
      ["Compile score (to file)"          cscsd-compile-score (null (cscsd-buffer-is-a-csd-p))]
      ["Play/Edit"           cscsd-play (cscsd-audio-available-p)])))

;  electric return

(define-key csound-sco-mode-map "\r" 'cscsd-sco-electric-return)

(defvar cscsd-sco-has-electric-return nil
  "")
(make-local-variable 'cscsd-sco-has-electric-return)

(defun cscsd-sco-electric-return ()
  ""
  (interactive)
  (if (and cscsd-sco-has-electric-return
           (scomx-i-statement-p)
           (numberp (read (match-string 2)))
           (numberp (read (match-string 3))))
      (progn
        (end-of-line)
        (insert "\ni " (match-string 1) 
                (format " %s" (+ (read (match-string 2))
                                 (read (match-string 3))))
                " " (match-string 3))
        (when (match-string 4)
          (insert (match-string 4))))
    (insert "\n")))

;;======================================================================

(provide 'csound-sco)
(run-hooks 'csound-sco-mode-load-hook)

;;; csound-sco.el ends here
